//********************************************************************************
//
// Lucide GBM plug-in
//
// Copyright (C) 2006 Heiko Nitzsche
//
//   This software is provided 'as-is', without any express or implied
//   warranty.  In no event will the author be held liable for any damages
//   arising from the use of this software.
//
//   Permission is granted to anyone to use this software for any purpose,
//   including commercial applications, and to alter it and redistribute it
//   freely, subject to the following restrictions:
//
//   1. The origin of this software must not be misrepresented; you must not
//      claim that you wrote the original software. If you use this software
//      in a product, an acknowledgment in the product documentation would be
//      appreciated but is not required.
//   2. Altered source versions must be plainly marked as such, and must not be
//      misrepresented as being the original software.
//   3. This notice may not be removed or altered from any source distribution.
//
// Date   : 16-Jul-2006
// Version: 1.0
//
//********************************************************************************
//
// Supported feature:
// ==================
// - document info
// - loading all formats supported by GBM
// - multipage bitmap viewing support
// - load bitmap data on demand (requested by the user)
// - scaled  rendering to a buffer
// - rotated rendering to a buffer
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
//********************************************************************************
//
// History: 27-Jun-2006: Initial public beta
//          30-Jun-2006: Public beta 2
//                       Fixes for the most urgent issues
//                       - file extension list too long for file dialog (only keep the first extension)
//                       - Add error message reporting to load_file())
//          16-Jul-2006: Added document info
//                       Added rotated rendering
//                       Added delayed loading of bitmap data (on user request)
//                       Support transparent caching for fast scrolling of a scaled view
//
//********************************************************************************

/*
 *  This file was generated by the SOM Compiler.
 *  Generated using:
 *     SOM incremental update: 2.24
 */

#ifdef __IBMCPP__
  #pragma strings( readonly )
#endif

#include "lugbm.hpp"

#ifndef SOM_Module_lugbm_Source
#define SOM_Module_lugbm_Source
#endif
#define LuGbmDocument_Class_Source

#include "lugbm.xih"

#include <stdio.h>
#include <string.h>

#include "os2defs.h"
#include "GbmDocument.hpp"
#include "gbm.h"

// ---------------------------------------------------------

// set to TRUE  if rendering should be done to a buffer (always 24bpp)
// set to FALSE if rendering should be done to HPS      (native bitmap bpp)
// (rendering to HPS is not yet ready for prime time)
#define RENDER_TO_BUFFER  FALSE

#define PLUGIN_VERSION    "1.00"

// ---------------------------------------------------------

static char * supportedExtensions = NULL;
static char * pluginDescription   = NULL;

// ---------------------------------------------------------

// Init / Terminate DLL

static unsigned libInitialize(unsigned termination)
{
    if ( termination )
    {
        /* DLL is detaching from process */
        free(supportedExtensions);
        supportedExtensions = NULL;

        delete [] pluginDescription;
        pluginDescription = NULL;
    }
    else
    {
        /* DLL is attaching to process */

        /* get all supported extensions from GBM */
        Gbm_init();

        int n_ft = 0;
        Gbm_query_n_filetypes(&n_ft);

        // we assume here that none of the import filters has more than 10 extensions with 3 chars
        int extensions_length = 10 * 3 * n_ft + 1;

        supportedExtensions = (char *) malloc(extensions_length);
        if (supportedExtensions == NULL)
        {
            Gbm_deinit();
            return 0;
        }
        strcpy(supportedExtensions, "");

        for (int ft = 0; ft < n_ft; ft++)
        {
            GBMFT gbmft;

            Gbm_query_filetype(ft, &gbmft);

            // get all extensions
            char * extTok = strtok(gbmft.extensions, " ");

#if 1 // Limit the number of reported extension to only the first one (most often used one)
      // until the file dialog extension list issue with the limited filename entry field
      // length is solved.

            if (extTok != NULL)
            {
                BOOLEAN tokenAdded = FALSE;

                // ignore IAX as for it the user would have to specify the size explicitly (no header))
                if (strcmp(extTok, "IAX") != 0)
                {
                    strcat(supportedExtensions, extTok);
                    tokenAdded = TRUE;
                }
                if ((ft + 1 < n_ft) && tokenAdded)
                {
                    strcat(supportedExtensions, ";");
                }
            }
#else
            while (extTok != NULL)
            {
                BOOLEAN tokenAdded = FALSE;

                if (strlen(supportedExtensions) + strlen(extTok) + 1 > extensions_length)
                {
                    // resize buffer
                    extensions_length += 10;
                    supportedExtensions = (char *) realloc(supportedExtensions, extensions_length);
                    if (supportedExtensions == NULL)
                    {
                        Gbm_deinit();
                        return 0;
                    }
                }
                // ignore IAX as for it the user would have to specify the size explicitly (no header))
                if (strcmp(extTok, "IAX") != 0)
                {
                    strcat(supportedExtensions, extTok);
                    tokenAdded = TRUE;
                }
                extTok = strtok(NULL, " ");
                if (tokenAdded)
                {
                    if (ft + 1 < n_ft)
                    {
                        strcat(supportedExtensions, ";");
                    }
                    else
                    {
                        if (extTok != NULL)
                        {
                            strcat(supportedExtensions, ";");
                        }
                    }
                }
            }
#endif
        }

        // create the plugin description string
        const char * descFormat = "GBM plugin ("PLUGIN_VERSION"), based on Generalised Bitmap Module (GBM.DLL: %.2f)";

        pluginDescription = new char[strlen(descFormat) + 20 + 1];
        sprintf(pluginDescription, descFormat, ((float)Gbm_version())/100.0);

        Gbm_deinit();
    }

    return( 1 );
}

// ---------------------------------------------------------

#ifdef __IBMCPP__

/* _CRT_init is the C run-time environment initialization function.         */
/* It will return 0 to indicate success and -1 to indicate failure.         */
extern "C" int _CRT_init(void);

/* _CRT_term is the C run-time environment termination function.            */
/* It only needs to be called when the C run-time functions are statically  */
/* linked.                                                                  */
extern "C" void _CRT_term(void);

/* __ctordtorInit is the C++ run-time environment initialization function.  */
/* It is required to initialize static objects and destructors              */
extern "C" void __ctordtorInit(void);

/* __ctordtorTerm is the C++ run-time environment termination function.  */
/* It is required to destroy static objects and destructors              */
extern "C" void __ctordtorTerm(void);

/* _DLL_InitTerm() - called by the loader for DLL initialization/termination  */
/* This function must return a non-zero value if successful and a zero value  */
/* if unsuccessful.                                                           */

extern "C" unsigned long _System _DLL_InitTerm( unsigned long hModule, unsigned long termination )
{
   switch(termination)
   {
       case 0:
           if (_CRT_init() == -1)
           {
               return 0;
           }
           __ctordtorInit();

         // init our DLL
         if (libInitialize(0) == 0)
         {
             return 0;
         }
         break;

      case 1:
         // terminate our DLL
         if (libInitialize(1) == 0)
         {
             return 0;
         }

         __ctordtorTerm();
         _CRT_term();
         break;

      default:
         return 0;
   }

   /* Indicate success.  Non-zero means success!!! */
   return 1;
}

#endif // ifdef __IBMCPP__

// -------------------------------------------------------
// -------------------------------------------------------

#ifdef __WATCOM_CPLUSPLUS__

extern "C" unsigned _System LibMain( unsigned hmod, unsigned termination )
{
    if ( termination )
    {
        // terminate our DLL
        if (libInitialize(1) == 0)
        {
            return 0;
        }
    }
    else
    {
        /* DLL is attaching to process */

        // init our DLL
        if (libInitialize(0) == 0)
        {
            return 0;
        }
    }

    return( 1 );
}
#endif // ifdef __WATCOM_CPLUSPLUS__

// ---------------------------------------------------------

extern "C" LuDocument * _System createObject()
{
    return new LuGbmDocument();
}

// ---------------------------------------------------------

extern "C" char * _System getSupportedExtensions()
{
    return supportedExtensions;
}

// ---------------------------------------------------------

extern "C" char * _System getDescription()
{
    return pluginDescription;
}

// ---------------------------------------------------------
// ---------------------------------------------------------

/**
 * loadFile
 * filename: filename of the file to load
 * password: password to unlock the file with, or NULL
 * error: Return location for an error, or NULL
 *
 * Loads the specified file.
 *
 * Return value: TRUE if file loaded, FALSE otherwise.
 *
 * Note: error string allocated using SOMMalloc(),
 *       use SOMFree() to free.
 **/

SOM_Scope boolean  SOMLINK loadFile(LuGbmDocument *somSelf ,
                                    Environment   *ev      , string  filename,
                                    string         password, string* error)
{
    LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    GbmDocument       *d       = (GbmDocument *)somThis->data;

    try
    {
        d->setDocumentFile(filename, NULL             /* guess format from file/extension */,
                                     TRUE             /* allow delayed loading of page data */,
                                     RENDER_TO_BUFFER /* 24bpp if rendering to a buffer */);
    }
    catch(GbmException & ex)
    {
        const char * errorMessage       = ex.getErrorMessage();
        const int    errorMessageLength = strlen(errorMessage);

        *error = (char *) SOMMalloc(errorMessageLength + 1);
        strcpy(*error, errorMessage);

        return FALSE;
    }

    return TRUE;
}

// ---------------------------------------------------------

/**
 * getBpp
 *
 * Return value: Number of bytes per pixel used for draw.
 **/

SOM_Scope short  SOMLINK getBpp(LuGbmDocument *somSelf, Environment *ev)
{
    /* if rendering to a buffer, always use 24bpp */
    return (RENDER_TO_BUFFER ? 3 : 0); // save memory if no buffer is used
}

// ---------------------------------------------------------

/**
 * isScalable
 *
 * (See renderPageToPixbuf(), scale)
 *
 * Return value: TRUE if document scalable, FALSE otherwise.
 **/

SOM_Scope boolean  SOMLINK isScalable(LuGbmDocument *somSelf,
                                      Environment   *ev)
{
    return TRUE;
}

// ---------------------------------------------------------

/**
 * isRotable
 *
 * (See renderPageToPixbuf(), rotation)
 *
 * Return value: TRUE if document can be rotated, FALSE otherwise.
 **/

SOM_Scope boolean  SOMLINK isRotable(LuGbmDocument *somSelf,
                                     Environment   *ev)
{
    return TRUE;
}

// ---------------------------------------------------------

/**
 * isRenderIntoPS
 *
 * See renderPageToPS()
 *
 * Return value: TRUE if document will render directly into
 *               presentation space.
 **/
SOM_Scope boolean  SOMLINK isRenderIntoPS(LuGbmDocument *somSelf,
                                          Environment   *ev)
{
    return (RENDER_TO_BUFFER ? FALSE : TRUE);
}

// ---------------------------------------------------------

/**
 * getPageCount
 *
 * Returns the number of pages in a loaded document.
 *
 * Return value: Number of pages
 **/

SOM_Scope long  SOMLINK getPageCount(LuGbmDocument *somSelf,
                                     Environment   *ev)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    const GbmDocument       *d       = (GbmDocument *)somThis->data;

    return d->getNumberOfPages();
}

// ---------------------------------------------------------

/**
 * getPageSize
 * pagenum: Page number
 * width: return location for the width of page
 *        (NULL if width not needed)
 * height: return location for the height of page
 *         (NULL if height not needed)
 *
 * Gets the size of specified page.
 **/

SOM_Scope void  SOMLINK getPageSize(LuGbmDocument *somSelf,
                                    Environment   *ev     , long     pagenum,
                                    double        *width  , double * height)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
          GbmDocument       *d       = (GbmDocument *)somThis->data;

    try
    {
        GbmDocumentPage & page(d->getPage(pagenum));

        if (page.getRotation() % 180)
        {
            *width  = page.getHeight();
            *height = page.getWidth();
        }
        else
        {
            *width  = page.getWidth();
            *height = page.getHeight();
        }
    }
    catch(GbmException & ex)
    {
        *width  = 0;
        *height = 0;
        return;
    }
}

// ---------------------------------------------------------

/**
 * getDocumentInfo
 *
 * Return value: a new LuDocumentInfo.
 **/

SOM_Scope LuDocumentInfo*  SOMLINK getDocumentInfo(LuGbmDocument *somSelf,
                                                   Environment   *ev)
{
    LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    GbmDocument       *d       = (GbmDocument *)somThis->data;

    LuDocumentInfo *info = (LuDocumentInfo *)SOMMalloc( sizeof( LuDocumentInfo ) );
    memset( info, 0, sizeof( LuDocumentInfo ) );

    // set long format description
    const char * formatDescription = d->getLongDescription();
    info->format = (char *) SOMMalloc( strlen(formatDescription) + 1 );
    strcpy(info->format, formatDescription);
    info->fields_mask |= LU_DOCUMENT_INFO_FORMAT;

    // set number of pages
    info->n_pages = d->getNumberOfPages();
    info->fields_mask |= LU_DOCUMENT_INFO_N_PAGES;

    // set layout to single page (any effect?)
    //info->layout = LU_DOCUMENT_LAYOUT_SINGLE_PAGE;
    //info->fields_mask |= LU_DOCUMENT_INFO_LAYOUT;

    return info;
}

// ---------------------------------------------------------

/**
 * renderPageToPixbuf
 * pagenum: the page to render from
 * src_x: x coordinate of upper left corner
 * src_y: y coordinate of upper left corner
 * src_width: width of rectangle to render
 * src_height: height of rectangle to render
 * scale: scale specified as pixels per point
 *        (if isScalable() is FALSE, scale ignored and assumed 1.0)
 * rotation: rotate the document by the specified degree
 *        (if isRotable() is FALSE, rotation ignored and assumed 0)
 * pixbuf: pixbuf to render into
 *
 * First scale the document to match the specified pixels per point,
 * then render the rectangle given by the upper left corner at
 * (src_x, src_y) and src_width and src_height.
 **/

SOM_Scope void  SOMLINK renderPageToPixbuf(LuGbmDocument *somSelf,
                                           Environment   *ev,
                                           long           pagenum   , long       src_x,
                                           long           src_y     , long       src_width,
                                           long           src_height, double     scale,
                                           long           rotation  , LuPixbuf * pixbuf)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
          GbmDocument       *d       = (GbmDocument *)somThis->data;

    try
    {
        if ((pixbuf == NULL) || (pagenum < 0) || (pagenum >= d->getNumberOfPages()))
        {
            // show a gray image in case of an error
            memset(pixbuf->getDataPtr(ev), 0x80, pixbuf->getRowSize(ev) * pixbuf->getHeight(ev));
            return;
        }

        GbmDocumentPage & page(d->getPage(pagenum));
        GbmRenderer     & renderer(d->getRenderer());

        renderer.setRenderPage(&page);
        renderer.renderToBuffer(src_x    , src_y,
                                src_width, src_height,
                                scale    , scale     , rotation,
                                (byte *) pixbuf->getDataPtr(ev));
    }
    catch(GbmException & ex)
    {
        // show a gray image in case of an error
        memset(pixbuf->getDataPtr(ev), 0x80, pixbuf->getRowSize(ev) * pixbuf->getHeight(ev));
    }
}

// ---------------------------------------------------------

/**
 * renderPageToPS
 * pagenum: the page to render from
 * src_x: x coordinate of upper left corner
 * src_y: y coordinate of upper left corner
 * src_width: width of rectangle to render
 * src_height: height of rectangle to render
 * scale: scale specified as pixels per point
 *        (if isScalable() is FALSE, scale ignored and assumed 1.0)
 * rotation: rotate the document by the specified degree
 *        (if isRotable() is FALSE, rotation ignored and assumed 0)
 * hps: handle of presentation space to render into
 * rect: pointer to RECTL structure, defines render area on HPS
 *
 * First scale the document to match the specified pixels per point,
 * then render the rectangle given by the upper left corner at
 * (src_x, src_y) and src_width and src_height.
 **/

SOM_Scope void  SOMLINK renderPageToPS(LuGbmDocument *somSelf  ,
                                       Environment   *ev       , long pagenum,
                                       long           src_x    , long src_y,
                                       long           src_width, long src_height,
                                       double         scale    , long rotation,
                                       unsigned long  hps      , somMToken rect)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
          GbmDocument       *d       = (GbmDocument *)somThis->data;

    const PRECTL renderArea = (const PRECTL) rect;

    if ((scale < 0.0) || (pagenum < 0) ||
        (src_y < 0)   || (src_x < 0)   || (src_width <= 0) || (src_height <= 0))
    {
        return;
    }

    try
    {
        if ((pagenum < 0) || (pagenum >= d->getNumberOfPages()))
        {
            return;
        }

        GbmDocumentPage & page(d->getPage(pagenum));
        GbmRenderer     & renderer(d->getRenderer());

        // allocate a render buffer
        const int drawStride = (src_width * page.getColorDepth() + 31)/31 * 4;
        byte * drawBuffer    = NULL;
        try
        {
            drawBuffer = new byte[drawStride * src_height];
        }
        catch(...)
        {
        }
        if (drawBuffer == NULL)
        {
            DosBeep(500,50);
            return;
        }

        renderer.setRenderPage(&page);
        renderer.renderToBuffer(src_x    , src_y,
                                src_width, src_height,
                                scale    , scale     , rotation,
                                drawBuffer);

        // Draw the bitmap
        // ===============

        // We have to create a modified local copy of the bitmap header
        // as the page header does not contain scaling and clipping info.
        #pragma pack(2)
        typedef struct
        {
           BITMAPINFOHEADER2 bmp2;
           RGB2 argb2Color[0x100];
        } BMP8HEADER;

        /* for above 8 bit images */
        typedef struct
        {
           BITMAPINFOHEADER2 bmp2;
        } BMP24HEADER;
        #pragma pack()


        BITMAPINFO2 * localHeader = NULL;
        if (page.getColorDepth() <= 8)
        {
            // bitmap contains a color palette
            localHeader = (BITMAPINFO2 *) new BMP8HEADER();
            memcpy(localHeader, page.getBitmapInfoPtr(), sizeof(BMP8HEADER));
        }
        else
        {
            // bitmap contains a color palette
            localHeader = (BITMAPINFO2 *) new BMP24HEADER();
            memcpy(localHeader, page.getBitmapInfoPtr(), sizeof(BMP24HEADER));
        }

        /* update relevant bitmap header data */
        localHeader->cx        = src_width;
        localHeader->cy        = src_height;
        localHeader->cBitCount = page.getColorDepth();
        localHeader->cbImage   = drawStride * src_height;

        // set forground/background color if we render a 1bpp bitmap
        if (page.getColorDepth() == 1)
        {
            const RGB2 * pRGBA = localHeader->argbColor;

            LONG colorBg = (pRGBA[0].bRed << 16) + (pRGBA[0].bGreen << 8) + pRGBA[0].bBlue;
            LONG colorFg = (pRGBA[1].bRed << 16) + (pRGBA[1].bGreen << 8) + pRGBA[1].bBlue;

            LONG colorTable[2] = { colorBg, colorFg };

            GpiCreateLogColorTable(hps, 0, LCOLF_CONSECRGB, 0, 2, (PLONG) &colorTable);
            /* if something goes wrong when creating color table, use the closest color */
            GpiSetBackColor(hps, GpiQueryColorIndex(hps, 0, colorBg));
            GpiSetColor    (hps, GpiQueryColorIndex(hps, 0, colorFg));
        }

        RECTL aptlPoints[2] = { 0 };

        aptlPoints[0] = *renderArea;

        aptlPoints[1].xLeft   = 0;
        aptlPoints[1].yBottom = 0;
        aptlPoints[1].xRight  = src_width;
        aptlPoints[1].yTop    = src_height;

        if (GpiDrawBits( hps,
                         (PBYTE) drawBuffer,
                         localHeader,
                         4L,
                         (PPOINTL) aptlPoints,
                         ROP_SRCCOPY,
                         BBO_IGNORE ) == GPI_ERROR)
        {
            DosBeep(500,50);
        }

        delete localHeader;
        localHeader = NULL;

        delete [] drawBuffer;
        drawBuffer = NULL;
    }
    catch(GbmException & ex)
    {
        DosBeep(500,50);
    }
}

// ---------------------------------------------------------
// ---------------------------------------------------------

SOM_Scope void SOMLINK somDefaultInit(LuGbmDocument * somSelf,
                                      som3InitCtrl  * ctrl)
{
    LuGbmDocumentData *somThis;
    somInitCtrl        globalCtrl;
    somBooleanVector   myMask;
    LuGbmDocument_BeginInitializer_somDefaultInit;
    LuGbmDocument_Init_LuDocument_somDefaultInit(somSelf, ctrl);

    // local LuGbmDocument initialization code
    GbmDocument *d = new GbmDocument();
    somThis->data = d;
}

// ---------------------------------------------------------

SOM_Scope void SOMLINK somDestruct(LuGbmDocument *somSelf, octet doFree,
                                   som3DestructCtrl* ctrl)
{
    LuGbmDocumentData *somThis;
    somDestructCtrl    globalCtrl;
    somBooleanVector   myMask;
    LuGbmDocument_BeginDestructor;

    // local LuGbmDocument deinitialization code
    GbmDocument *d = (GbmDocument *)somThis->data;

    delete d;
    d = NULL;

    // end of local LuGbmDocument deinitialization code

    LuGbmDocument_EndDestructor;
}


