//********************************************************************************
// Mozilla GBM plug-in: npgbm.cpp
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
// Copyright (C) 2006 Heiko Nitzsche
//
//   This software is provided 'as-is', without any express or implied
//   warranty.  In no event will the author be held liable for any damages
//   arising from the use of this software.
//
//   Permission is granted to anyone to use this software for any purpose,
//   including commercial applications, and to alter it and redistribute it
//   freely, subject to the following restrictions:
//
//   1. The origin of this software must not be misrepresented; you must not
//      claim that you wrote the original software. If you use this software
//      in a product, an acknowledgment in the product documentation would be
//      appreciated but is not required.
//   2. Altered source versions must be plainly marked as such, and must not be
//      misrepresented as being the original software.
//   3. This notice may not be removed or altered from any source distribution.
//
// Date   : 16-Jul-2006
// Version: 1.0
//
//********************************************************************************
//
// Supported feature:
// ==================
// - loading all formats supported by GBM
// - scaled  rendering to a buffer
// - multipage preview possible once a GUI controller has been implemented
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
//********************************************************************************
//
// History: 16-Jul-2006:
//
//********************************************************************************

#ifdef __IBMCPP__
  #pragma strings( readonly )
#endif

#include <string.h>
#include <stdlib.h>

#include "os2defs.h"
#include "GbmDocument.hpp"
#include "gbm.h"

#ifndef _NPAPI_H_
#include "npapi.h"
#endif

//
// Instance state information about the plugin.
//

typedef struct _PluginInstance PluginInstance;
typedef struct _PluginInstance
{
    NPWindow       * fWindow;
    HWND             hWnd;
    uint16           fMode;
    PFNWP            lpfnOldWndProc;
    NPSavedData    * pSavedInstanceData;
    GbmDocument    * pDocument;
    int              gbmTypeIndex;
    int              progress;
    int              progressMax;
    PluginInstance * pNext;
} PluginInstance;

// ---------------------------------------------------------

typedef struct LOOKUP_TABLE_DEF
{
    const char * mimeType;
    const char * extension;
};

static struct LOOKUP_TABLE_DEF LOOKUP_TABLE [] =
{
    { "image/bmp"               , "BMP" },
    { "image/x-windows-bmp"     , "BMP" },
    { "image/gif"               , "GIF" },
    { "image/x-pcx"             , "PCX" },
    { "image/tiff"              , "TIF" },
    { "image/x-tiff"            , "TIF" },
    { "image/tga"               , "TGA" },
    { "image/iff"               , "IFF" },
    { "image/vid"               , "VID" },
    { "image/x-portable-bitmap" , "PBM" },
    { "image/x-portable-graymap", "PGM" },
    { "image/x-portable-greymap", "PGM" },
    { "image/x-portable-pixmap" , "PPM" },
    { "image/x-portable-anymap" , "PNM" },
    { "image/png"               , "PNG" },
    { "image/kps"               , "KPS" },
    { "image/x-xbm"             , "XBM" },
    { "image/xbm"               , "XBM" },
    { "image/x-xbitmap"         , "XBM" },
    { "image/spr"               , "SPR" },
    { "image/pseg"              , "PSE" },
    { "image/img"               , "IMG" },
    { "image/cvp"               , "CVP" },
    { "image/jpeg"              , "JPG" },
    { "image/pjpeg"             , "JPG" }
};

const int LOOKUP_TABLE_LENGTH = sizeof(LOOKUP_TABLE) /
                                sizeof(LOOKUP_TABLE[0]);

// ---------------------------------------------------------

MRESULT APIENTRY SubClassFunc(HWND hWnd,ULONG Message,MPARAM wParam, MPARAM lParam);

void Draw(PluginInstance *This, HPS hps, RECTL *rect);

HMODULE DLLInstance;

// ---------------------------------------------------------

//
// This plugin will always be linked statically.
// Thus only the DLL init/term for this mode is implemented.
//

// ---------------------------------------------------------

#ifdef __IBMCPP__

/* _CRT_init is the C run-time environment initialization function.         */
/* It will return 0 to indicate success and -1 to indicate failure.         */
extern "C" int _CRT_init(void);

/* _CRT_term is the C run-time environment termination function.            */
/* It only needs to be called when the C run-time functions are statically  */
/* linked.                                                                  */
extern "C" void _CRT_term(void);

/* __ctordtorInit is the C++ run-time environment initialization function.  */
/* It is required to initialize static objects and destructors              */
extern "C" void __ctordtorInit(void);

/* __ctordtorTerm is the C++ run-time environment termination function.  */
/* It is required to destroy static objects and destructors              */
extern "C" void __ctordtorTerm(void);

/* _DLL_InitTerm() - called by the loader for DLL initialization/termination  */
/* This function must return a non-zero value if successful and a zero value  */
/* if unsuccessful.                                                           */

extern "C" unsigned long _System _DLL_InitTerm( unsigned long hModule, unsigned long termination )
{
   DLLInstance = (HMODULE) hModule;

   switch(termination)
   {
       case 0:
          if (_CRT_init() == -1)
          {
              return 0;
          }
          __ctordtorInit();
          break;

      case 1:
          __ctordtorTerm();
          _CRT_term();
          break;

      default:
          return 0;
   }

   /* Indicate success.  Non-zero means success!!! */
   return 1;
}

#endif // ifdef __IBMCPP__

// -------------------------------------------------------
// -------------------------------------------------------

#ifdef __WATCOM_CPLUSPLUS__

extern "C" unsigned _System LibMain( unsigned hmod, unsigned termination )
{
    DLLInstance = (HMODULE) hmod;

    if ( termination )
    {
    }
    else
    {
        /* DLL is attaching to process */
    }

    return( 1 );
}
#endif // ifdef __WATCOM_CPLUSPLUS__

// -------------------------------------------------------
// -------------------------------------------------------

// A plugin instance typically will subclass the plugin's client window, so
// it can get Windows messages, (such as paint, palettechanged, keybd, etc).
// To do work associated with a specific plugin instance the WndProc which
// receives the Windows messages, (named "SubClassFunc" herein), needs access
// to the "This" (PluginInstance*) ptr.

//* OS2 bugbug -- we need to reserve 4 bytes of window words to correct this
//  gorp.....

// When Navigator registers the plugin client's window class, (the class for
// the window passed in NPP_SetWindow()), Navigator does not reserve any
// "extra" windows bytes.  If it had, the plugin could simply have stored its
// "This" (PluginInstance*) ptr in the extra bytes.  But Nav did not, and the
// plugin cannot, so a different technique must be used.  The technique used
// is to keep a linked list of PluginInstance structures, and walk the list
// to find which one is associated with the window handle.  Inefficient,
// grungy, complicates the code, etc.  C'est la vie ...

PluginInstance* g_pHeadInstanceList = 0;

// Associate the hWnd with pInstance by setting the hWnd member of the
// PluginInstance struct.  Also, add the PluginInstance struct to the list
// if necessary
static void AssociateInstance(HWND hWnd, PluginInstance* pInstance)
{
    pInstance->hWnd = hWnd;     // redundant, but usefull to get hwnd from
                                // pinstance later.
    BOOL rc = WinSetWindowULong(hWnd, QWL_USER, (ULONG)pInstance);
}

// Find the PluginInstance associated with this hWnd and return it
static PluginInstance* GetInstance(HWND hWnd)
{
    return (PluginInstance*)WinQueryWindowULong(hWnd, QWL_USER);
}

//----------------------------------------------------------------------------
// NPP_Initialize:
//----------------------------------------------------------------------------
NPError NPP_Initialize(void)
{
    // do your one time initialization here, such as dynamically loading
    // dependant DLLs
    return NPERR_NO_ERROR;
}


//----------------------------------------------------------------------------
// NPP_Shutdown:
//----------------------------------------------------------------------------
void NPP_Shutdown(void)
{
    // do your one time uninitialization here, such as unloading dynamically
    // loaded DLLs
}


//----------------------------------------------------------------------------
// NPP_New:
//----------------------------------------------------------------------------
NPError NP_LOADDS NPP_New(NPMIMEType    pluginType,
                          NPP           instance,
                          uint16        mode,
                          int16         argc,
                          char        * argn[],
                          char        * argv[],
                          NPSavedData * saved)
{
    if (instance == NULL)
    {
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    // lookup the file extension valid for the mime type
    int gbmTypeIndex = -1;

    if (pluginType != NULL)
    {
        for (int index = 0; index < LOOKUP_TABLE_LENGTH; index++)
        {
            if (strcmp(LOOKUP_TABLE[index].mimeType, pluginType) == 0)
            {
                gbmTypeIndex = index;
                break;
            }
        }
    }

    // don't try to catch an unknown type even though it might work in some cases
    if (gbmTypeIndex == -1)
    {
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
    PluginInstance* This = (PluginInstance*) instance->pdata;

    if (This == NULL)
    {
        return NPERR_OUT_OF_MEMORY_ERROR;
    }

    //
    // *Developers*: Initialize fields of your plugin
    // instance data here.  If the NPSavedData is non-
    // NULL, you can use that data (returned by you from
    // NPP_Destroy to set up the new plugin instance.
    //

    This->fWindow            = 0;
    // mode is NP_EMBED, NP_FULL, or NP_BACKGROUND (see npapi.h)
    This->fMode              = mode;
    This->hWnd               = 0;
    This->pSavedInstanceData = saved;
    This->pNext              = NULL;

    This->pDocument          = new GbmDocument();
    This->gbmTypeIndex       = gbmTypeIndex;
    This->progress           = 0;
    This->progressMax        = 0;

    return NPERR_NO_ERROR;
}


//-----------------------------------------------------------------------------
// NPP_Destroy:
//-----------------------------------------------------------------------------
NPError NP_LOADDS NPP_Destroy(NPP instance, NPSavedData** save)
{
    if (instance == 0)
    {
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    PluginInstance* This = (PluginInstance*) instance->pdata;

    //
    // *Developers*: If desired, call NP_MemAlloc to create a
    // NPSavedDate structure containing any state information
    // that you want restored if this plugin instance is later
    // recreated.
    //

    if (This != 0)
    {
        delete This->pDocument;
        This->pDocument = NULL;

        This->gbmTypeIndex = -1;
        This->progress     = 0;
        This->progressMax  = 0;

        // Remove the subclass for the client window
        if(This->hWnd)
        {
            WinSubclassWindow(This->hWnd, This->lpfnOldWndProc);
        }

        // make some saved instance data if necessary
        if (This->pSavedInstanceData == 0)
        {
            // make a struct header for the data
            This->pSavedInstanceData =
                (NPSavedData*)NPN_MemAlloc(sizeof (struct _NPSavedData));

            // fill in the struct
            if(This->pSavedInstanceData != 0)
            {
                This->pSavedInstanceData->len = 0;
                This->pSavedInstanceData->buf = 0;

                // replace the def below and references to it with your data
                #define SIDATA "aSavedInstanceDataBlock"

                // the data
                This->pSavedInstanceData->buf = NPN_MemAlloc(sizeof SIDATA);

                if(This->pSavedInstanceData->buf != 0)
                {
                    strcpy((char*)This->pSavedInstanceData->buf, SIDATA);
                    This->pSavedInstanceData->len = sizeof SIDATA;
                }
            }
        }

        // save some instance data
        *save = This->pSavedInstanceData;

        NPN_MemFree(instance->pdata);
        instance->pdata = 0;
    }

    return NPERR_NO_ERROR;
}


//----------------------------------------------------------------------------
// NPP_SetWindow:
//----------------------------------------------------------------------------
NPError NP_LOADDS NPP_SetWindow(NPP instance, NPWindow* window)
{
    if (instance == 0)
    {
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    PluginInstance* This = (PluginInstance*) instance->pdata;

    //
    // *Developers*: Before setting fWindow to point to the
    // new window, you may wish to compare the new window
    // info to the previous window (if any) to note window
    // size changes, etc.
    //
    if((window->window != 0) && (This->hWnd == 0))
    {
        This->fWindow = window;
        This->hWnd    = (HWND)This->fWindow->window;

        // subclass the window
        This->lpfnOldWndProc = WinSubclassWindow(This->hWnd, SubClassFunc);
        AssociateInstance(This->hWnd, This);
    }
    else
    {
        // if window handle changed
        if(This->hWnd != (HWND)window->window)
        {
            // remember the new window
            This->fWindow = window;

            // Remove the subclass for the old client window
            WinSubclassWindow(This->hWnd, This->lpfnOldWndProc);

            // remember the new window handle
            This->hWnd = (HWND)This->fWindow->window;

            if(This->hWnd != 0)
            {
                // subclass the new one
                This->lpfnOldWndProc = WinSubclassWindow(This->hWnd,
                                                         SubClassFunc);
                AssociateInstance(This->hWnd, This);
            }
        }
    }

    return NPERR_NO_ERROR;
}


//----------------------------------------------------------------------------
// NPP_NewStream:
//----------------------------------------------------------------------------
NPError NP_LOADDS NPP_NewStream(NPP          instance,
                                NPMIMEType   type,
                                NPStream   * stream,
                                NPBool       seekable,
                                uint16     * stype)
{
    if (instance == 0)
    {
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    PluginInstance* This = (PluginInstance*) instance->pdata;

    // ask for the whole file
    // (we might add streaming later by writing our own GBM I/O functions)
    // So far we use the streaming calls to display the loading progress.
    *stype = NP_ASFILE;

    This->progress    = 0;
    This->progressMax = stream->end;

    // invalidate window to ensure a redraw
    WinInvalidateRect(This->hWnd, 0, TRUE);

    return NPERR_NO_ERROR;
}


//
// *Developers*:
// These next 2 functions are directly relevant in a plug-in which handles the
// data in a streaming manner.  If you want zero bytes because no buffer space
// is YET available, return 0.  As long as the stream has not been written
// to the plugin, Navigator will continue trying to send bytes.  If the plugin
// doesn't want them, just return some large number from NPP_WriteReady(), and
// ignore them in NPP_Write().  For a NP_ASFILE stream, they are still called
// but can safely be ignored using this strategy.
//

int32 STREAMBUFSIZE = 0X0FFFFFFF;   // If we are reading from a file in
                                    // NP_ASFILE mode, we can take any size
                                    // stream in our write call (since we
                                    // ignore it)

//----------------------------------------------------------------------------
// NPP_WriteReady:
//----------------------------------------------------------------------------
int32 NP_LOADDS NPP_WriteReady(NPP instance, NPStream *stream)
{
#if 0
    if (instance != 0)
    {
        PluginInstance* This = (PluginInstance*) instance->pdata;
    }
#endif

    return STREAMBUFSIZE;   // Number of bytes ready to accept in NPP_Write()
}


//----------------------------------------------------------------------------
// NPP_Write:
//----------------------------------------------------------------------------
int32 NP_LOADDS NPP_Write(NPP instance, NPStream *stream,
                          int32 offset, int32     len   , void *buffer)
{
    if (instance != 0)
    {
        PluginInstance* This = (PluginInstance*) instance->pdata;

        This->progress += len;

        // invalidate window to ensure a redraw
        WinInvalidateRect(This->hWnd, 0, TRUE);
    }

    return len;     // The number of bytes accepted.  Return a
                    // negative number here if, e.g., there was an error
                    // during plugin operation and you want to abort the
                    // stream
}


//----------------------------------------------------------------------------
// NPP_DestroyStream:
//----------------------------------------------------------------------------
NPError NP_LOADDS NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
    if (instance == 0)
    {
        return NPERR_INVALID_INSTANCE_ERROR;
    }
#if 0
    PluginInstance* This = (PluginInstance*) instance->pdata;
#endif
    return NPERR_NO_ERROR;
}


//----------------------------------------------------------------------------
// NPP_StreamAsFile:
//----------------------------------------------------------------------------
void NP_LOADDS NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
    if (instance == 0)
    {
        return;
    }

    PluginInstance* This = (PluginInstance*) instance->pdata;

    try
    {
        if (This->gbmTypeIndex == -1)
        {
            // set bitmap for use in drawing and printing (lookup format via filename)
            This->pDocument->setDocumentFile(fname, NULL,
                                                    FALSE /* force immediate loading of page data */,
                                                    FALSE /* allow also non 24bpp */);
        }
        else
        {
            // set bitmap for use in drawing and printing (lookup format via provided extension)
            This->pDocument->setDocumentFile(fname, LOOKUP_TABLE[This->gbmTypeIndex].extension,
                                                    FALSE /* force immediate loading of page data */,
                                                    FALSE /* allow also non 24bpp */);
        }
    }
    catch(...)
    {
    }

    // invalidate window to ensure a redraw
    WinInvalidateRect(This->hWnd, 0, TRUE);
}


//----------------------------------------------------------------------------
// NPP_Print:
//----------------------------------------------------------------------------
void NP_LOADDS NPP_Print(NPP instance, NPPrint* printInfo)
{
    if (printInfo == 0)   // trap invalid parm
    {
        return;
    }

    if (instance != 0)
    {
        PluginInstance* This = (PluginInstance*) instance->pdata;

        if (printInfo->mode == NP_FULL)
        {
            //
            // *Developers*: If your plugin would like to take over
            // printing completely when it is in full-screen mode,
            // set printInfo->pluginPrinted to TRUE and print your
            // plugin as you see fit.  If your plugin wants Netscape
            // to handle printing in this case, set printInfo->pluginPrinted
            // to FALSE (the default) and do nothing.  If you do want
            // to handle printing yourself, printOne is true if the
            // print button (as opposed to the print menu) was clicked.
            // On the Macintosh, platformPrint is a THPrint; on Windows,
            // platformPrint is a structure (defined in npapi.h) containing
            // the printer name, port, etc.
            //
            void* platformPrint = printInfo->print.fullPrint.platformPrint;
            NPBool printOne = printInfo->print.fullPrint.printOne;

            printInfo->print.fullPrint.pluginPrinted = FALSE; // Do the default
        }
        else    // If not fullscreen, we must be embedded
        {
            //
            // *Developers*: If your plugin is embedded, or is full-screen
            // but you returned false in pluginPrinted above, NPP_Print
            // will be called with mode == NP_EMBED.  The NPWindow
            // in the printInfo gives the location and dimensions of
            // the embedded plugin on the printed page.  On the Macintosh,
            // platformPrint is the printer port; on Windows, platformPrint
            // is the handle to the printing device context.
            //
            NPWindow* printWindow = &(printInfo->print.embedPrint.window);
            void* platformPrint = printInfo->print.embedPrint.platformPrint;

            // Heiko: The original code did not work and moved the elements totally
            //        outside the visible area. Thus it is changed (and simplified).

            /* get Presentation Space and save it */
            HPS hps = (HPS)platformPrint;

            SIZEL siz = { 0,0 };
            GpiQueryPS(hps, &siz);

            /* draw using common drawing routine */
            RECTL rect = { (int)printWindow->x,
                           (int)(siz.cy - (printWindow->y + printWindow->height)),
                           (int)(rect.xLeft   + printWindow->width),
                           (int)(rect.yBottom + printWindow->height)
                         };

            Draw(This, hps, &rect);
        }
    }
}


//----------------------------------------------------------------------------
// NPP_HandleEvent:
// Mac-only.
//----------------------------------------------------------------------------
int16 NP_LOADDS NPP_HandleEvent(NPP instance, void* event)
{
    NPBool eventHandled = FALSE;
    if (instance == 0)
    {
        return eventHandled;
    }

    PluginInstance* This = (PluginInstance*) instance->pdata;

    //
    // *Developers*: The "event" passed in is a Macintosh
    // EventRecord*.  The event.what field can be any of the
    // normal Mac event types, or one of the following additional
    // types defined in npapi.h: getFocusEvent, loseFocusEvent,
    // adjustCursorEvent.  The focus events inform your plugin
    // that it will become, or is no longer, the recepient of
    // key events.  If your plugin doesn't want to receive key
    // events, return false when passed at getFocusEvent.  The
    // adjustCursorEvent is passed repeatedly when the mouse is
    // over your plugin; if your plugin doesn't want to set the
    // cursor, return false.  Handle the standard Mac events as
    // normal.  The return value for all standard events is currently
    // ignored except for the key event: for key events, only return
    // true if your plugin has handled that particular key event.
    //

    return eventHandled;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

//
// Here is the subclass function.
//
MRESULT APIENTRY SubClassFunc(HWND   hWnd,
                              ULONG  Message,
                              MPARAM wParam,
                              MPARAM lParam)
{
    PluginInstance *This = GetInstance(hWnd);

    switch(Message)
    {
        case WM_REALIZEPALETTE:
            WinInvalidateRect(hWnd, 0, TRUE);
            WinUpdateWindow(hWnd);
            return 0;
            break;

        case WM_PAINT:
        {
            /* invalidate the whole window */
            // WinInvalidateRect(hWnd, NULL, TRUE);

            /* get PS associated with window and set to PU_PELS coordinates */
            RECTL invalidRect;
            HPS hps = WinBeginPaint(hWnd, NULL, &invalidRect);

            /* get window size */
            RECTL rect;
            WinQueryWindowRect(hWnd, &rect);

            /* draw using common drawing routine */
            Draw(This, hps, &rect);

            WinEndPaint(hps);
        }
        return (MRESULT)0;
        break;

        default:
            break;
    }

    return ((PFNWP)This->lpfnOldWndProc)(hWnd, Message, wParam, lParam);
}

//----------------------------------------------------------------------------

void Draw(PluginInstance *This, HPS hps, RECTL *rect)
{
    if (This->pDocument->getNumberOfPages() < 1)
    {
        POINTL startPoint = { rect->xLeft , rect->yBottom };
        POINTL endPoint   = { rect->xRight, rect->yTop    };

        if ((This->progress    > 0) &&
            (This->progressMax > 0) &&
            (This->progress    <= This->progressMax))
        {
            endPoint.y = rect->yTop * This->progress / This->progressMax;
            GpiSetColor(hps, CLR_DARKGRAY);
            GpiMove(hps, &startPoint);
            GpiBox(hps, DRO_OUTLINEFILL, &endPoint, 0L, 0L);

            startPoint.y = endPoint.y;
            endPoint.y   = rect->yTop;

            GpiSetColor(hps, CLR_PALEGRAY);
            GpiMove(hps, &startPoint);
            GpiBox(hps, DRO_OUTLINEFILL, &endPoint, 0L, 0L);
        }
        else
        {
            GpiSetColor(hps, CLR_PALEGRAY);
            GpiMove(hps, &startPoint);
            GpiBox(hps, DRO_OUTLINEFILL, &endPoint, 0L, 0L);
        }
        return;
    }

    byte * drawBuffer = NULL;

    try
    {
        const int src_x      = 0;
        const int src_y      = 0;
        const int src_width  = (int) (rect->xRight - rect->xLeft);
        const int src_height = (int) (rect->yTop   - rect->yBottom);

        // we show so far always page 0
        GbmDocumentPage & page(This->pDocument->getPage(0));
        GbmRenderer     & renderer(This->pDocument->getRenderer());

        const double scale_x = ((double) src_width) / page.getWidth();
        const double scale_y = ((double) src_height) / page.getHeight();
        const int rotation   = 0;

        // allocate a render buffer
        const int drawStride = (src_width * page.getColorDepth() + 31)/31 * 4;
        try
        {
            drawBuffer = new byte[drawStride * src_height];
        }
        catch(...)
        {
        }
        if (drawBuffer == NULL)
        {
            throw GbmException(GBM_ERR_MEM);
        }

        renderer.setRenderPage(&page);
        renderer.renderToBuffer(src_x    , src_y,
                                src_width, src_height,
                                scale_x  , scale_y   , rotation,
                                drawBuffer);

        // Draw the bitmap
        // ===============

        // We have to create a modified local copy of the bitmap header
        // as the page header does not contain scaling and clipping info.
        #pragma pack(2)
        typedef struct
        {
           BITMAPINFOHEADER2 bmp2;
           RGB2 argb2Color[0x100];
        } BMP8HEADER;

        /* for above 8 bit images */
        typedef struct
        {
           BITMAPINFOHEADER2 bmp2;
        } BMP24HEADER;
        #pragma pack()


        BITMAPINFO2 * localHeader = NULL;
        if (page.getColorDepth() <= 8)
        {
            // bitmap contains a color palette
            const BITMAPINFO2 * bmpInfo = page.getBitmapInfoPtr();
            if (bmpInfo == NULL)
            {
                throw GbmException(GBM_ERR_MEM);
            }
            localHeader = (BITMAPINFO2 *) new BMP8HEADER;
            if (localHeader == NULL)
            {
                throw GbmException(GBM_ERR_MEM);
            }
            memcpy(localHeader, bmpInfo, sizeof(BMP8HEADER));
        }
        else
        {
            // bitmap contains no color palette
            const BITMAPINFO2 * bmpInfo = page.getBitmapInfoPtr();
            if (bmpInfo == NULL)
            {
                throw GbmException(GBM_ERR_MEM);
            }
            localHeader = (BITMAPINFO2 *) new BMP24HEADER;
            if (localHeader == NULL)
            {
                throw GbmException(GBM_ERR_MEM);
            }
            memcpy(localHeader, bmpInfo, sizeof(BMP24HEADER));
        }

        /* update relevant bitmap header data */
        localHeader->cx        = src_width;
        localHeader->cy        = src_height;
        localHeader->cBitCount = page.getColorDepth();
        localHeader->cbImage   = drawStride * src_height;

        // set forground/background color if we render a 1bpp bitmap
        if (page.getColorDepth() == 1)
        {
            const RGB2 * pRGBA = localHeader->argbColor;

            LONG colorBg = (pRGBA[0].bRed << 16) + (pRGBA[0].bGreen << 8) + pRGBA[0].bBlue;
            LONG colorFg = (pRGBA[1].bRed << 16) + (pRGBA[1].bGreen << 8) + pRGBA[1].bBlue;

            LONG colorTable[2] = { colorBg, colorFg };

            GpiCreateLogColorTable(hps, 0, LCOLF_CONSECRGB, 0, 2, (PLONG) &colorTable);
            /* if something goes wrong when creating color table, use the closest color */
            GpiSetBackColor(hps, GpiQueryColorIndex(hps, 0, colorBg));
            GpiSetColor    (hps, GpiQueryColorIndex(hps, 0, colorFg));
        }

        RECTL aptlPoints[2] = { 0 };

        // target rect
        aptlPoints[0].xLeft   = rect->xLeft;
        aptlPoints[0].yBottom = rect->yBottom;
        aptlPoints[0].xRight  = rect->xRight;
        aptlPoints[0].yTop    = rect->yTop;

        // source rect (mirror vertically)
        aptlPoints[1].xLeft   = 0;
        aptlPoints[1].yBottom = src_height;
        aptlPoints[1].xRight  = src_width;
        aptlPoints[1].yTop    = 0;

        if (GpiDrawBits( hps,
                         (PBYTE) drawBuffer,
                         localHeader,
                         4L,
                         (PPOINTL) aptlPoints,
                         ROP_SRCCOPY,
                         BBO_IGNORE ) == GPI_ERROR)
        {
            POINTL startPoint = { rect->xLeft , rect->yBottom };
            POINTL endPoint   = { rect->xRight, rect->yTop    };

            GpiMove(hps, &startPoint);
            GpiSetColor(hps, CLR_PALEGRAY);
            GpiBox(hps, DRO_OUTLINEFILL, &endPoint, 0L, 0L);
        }

        delete localHeader;
        localHeader = NULL;

        delete [] drawBuffer;
        drawBuffer = NULL;
    }
    catch(...)
    {
        delete [] drawBuffer;
        drawBuffer = NULL;

        POINTL startPoint = { rect->xLeft , rect->yBottom };
        POINTL endPoint   = { rect->xRight, rect->yTop    };

        GpiMove(hps, &startPoint);
        GpiSetColor(hps, CLR_PALEGRAY);
        GpiBox(hps, DRO_OUTLINEFILL, &endPoint, 0L, 0L);
    }
}


